import logging
import click

from reactive_robot import PKG_DIR, PYTHON_VERSION, __version__
from reactive_robot.config.base import load_config
from reactive_robot.log import Log
from reactive_robot.serve import serve


def add_options(opts):
    def inner(f):
        for i in reversed(opts):
            f = i(f)
        return f

    return inner


def verbose_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(Log)
        if value:
            state.stream.setLevel(logging.DEBUG)

    return click.option('-v', '--verbose',
                        is_flag=True,
                        expose_value=False,
                        help='Enable verbose output',
                        callback=callback)(f)


def quiet_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(Log)
        if value:
            state.stream.setLevel(logging.ERROR)

    return click.option('-q', '--quiet',
                        is_flag=True,
                        expose_value=False,
                        help='Silence warnings',
                        callback=callback)(f)


common_options = add_options([quiet_option, verbose_option])
common_config_options = add_options([
    click.option('-f', '--config-file', type=click.File('rb'), help="Please provide a rx-robot config file."),
])


@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(
    __version__,
    '-V', '--version',
    message=f'%(prog)s, version %(version)s from {PKG_DIR} (Python {PYTHON_VERSION})'
)
@common_options
def cli():
    """
    RxRobot - Bind your robots with popular event streams (RabbitMQ, Kafka).
    """


@cli.command(name="serve")
@common_config_options
@common_options
def serve_robots(**kwargs):
    """Starts rx-robot event loop"""
    config = load_config(**kwargs)
    serve(config)



from marshmallow import Schema, fields, post_load

from reactive_robot.models import ReactiveRobotModel


class ConnectorSchema(Schema):
    driver = fields.String()
    connection_url = fields.String()


class RobotArgsSchema(Schema):
    pass


class RobotSchema(Schema):
    file = fields.Field()
    args = fields.Nested(RobotArgsSchema(), required=False, allow_none=True)


class BindingSchema(Schema):
    name = fields.String()
    topic = fields.String()
    robot = fields.Nested(RobotSchema(), allow_none=True)


class S3Schema(Schema):
    host = fields.Url()
    bucket_name = fields.String()
    access_key = fields.String()
    access_secret = fields.String()


class RxRobotConfigSchema(Schema):
    service_name = fields.String(required=True)
    service_version = fields.String()

    connector = fields.Nested(ConnectorSchema(), allow_none=True)
    bindings = fields.List(fields.Nested(BindingSchema()))

    @post_load
    def post_validation(self, data, **kwargs):
        return ReactiveRobotModel.from_dict(**data)



import logging
from abc import ABC
from typing import List
from urllib.parse import urlparse

from reactive_robot.parsers.string import RawPayloadParser

logger = logging.getLogger("reactive_robot.models")


class Model(ABC):

    def __init__(self, *args, **kwargs):
        pass


class RobotModel(Model):
    file: str
    args: str

    def __init__(self, *args, **kwargs):
        self.file = kwargs.pop("file")
        self.args = kwargs.pop("args")
        super(RobotModel, self).__init__(*args, **kwargs)


class ConnectorModel(Model):
    driver: str
    connection_url: str
    variable_parser = RawPayloadParser()

    def __init__(self, *args, **kwargs):
        self.driver = kwargs.pop("driver")
        self.connection_url = urlparse(kwargs.pop("connection_url"))
        super(ConnectorModel, self).__init__(*args, **kwargs)


class BindingModel(Model):
    name: str
    topic: str
    robot: RobotModel

    def __init__(self, *args, **kwargs):
        self.name = kwargs.pop("name")
        self.topic = kwargs.pop("topic")
        self.robot = RobotModel(**kwargs.pop("robot", {}))
        super(BindingModel, self).__init__(*args, **kwargs)


class ReactiveRobotModel(Model):
    service_name: str = "Reactive Robot Service"
    service_version: str = "0.1.0"

    connector: ConnectorModel = None
    bindings: List[BindingModel] = []

    def set_connector(self, connector: ConnectorModel):
        self.connector = connector

    def add_binding(self, binding: BindingModel):
        self.bindings.append(binding)

    @classmethod
    def from_dict(cls, **kwargs):
        model = ReactiveRobotModel()

        model.service_name = kwargs.pop("service_name")
        model.service_version = kwargs.pop("service_version")

        model.connector = ConnectorModel(**kwargs.pop("connector"))

        bindings = kwargs.pop("bindings", [])
        model.bindings = [BindingModel(**bd) for bd in bindings]

        logger.debug("Using configuration model, %s" % model)
        return model

    def __str__(self):
        return "<ReactiveRobotModel %s-%s>" % (self.service_name, self.service_version)



import subprocess
import tempfile
import pika

from typing import Iterable
from urllib.parse import ParseResult

from reactive_robot.connectors.base import Connector
from reactive_robot.models import BindingModel

import logging

logger = logging.getLogger("reactive_robot.connectors.rabbitmq")


class RabbitMQConnector(Connector):

    def _run_robot(self, topic, message):
        for b in self.bindings:
            if b.topic == topic:
                with tempfile.TemporaryDirectory() as tmpdirname:
                    cmd = " ".join(["robot",
                                    "--outputdir", tmpdirname,
                                    b.robot.file])

                    logger.info("Executing cmd, %s" % cmd)
                    result = subprocess.run(cmd.split(" "))
                    logger.info("Execution finished with return code %s" % result.returncode)

    def _map_binding(self, channel, binding: BindingModel):
        channel.exchange_declare(exchange=binding.topic, exchange_type='fanout')

        result = channel.queue_declare(queue='', exclusive=True)
        queue_name = result.method.queue

        channel.queue_bind(exchange=binding.topic, queue=queue_name)
        channel.basic_consume(queue=queue_name, on_message_callback=self.on_message, auto_ack=True)

    def on_message(self, ch, method, properties, body):
        logger.info("Received message %s from channel %s" % (body, method.exchange))
        self.executor.submit(lambda: self._run_robot(method.exchange, body))

    def bind(self, connection_url: ParseResult, bindings: Iterable[BindingModel]):
        connection = pika.BlockingConnection(pika.URLParameters(connection_url.geturl()))
        self.bindings = bindings

        channel = connection.channel()
        for binding in bindings:
            self._map_binding(channel, binding)

        channel.start_consuming()



import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable
from concurrent.futures.thread import ThreadPoolExecutor
from urllib.parse import ParseResult

from reactive_robot.constants import MAX_WORKER
from reactive_robot.models import BindingModel
from reactive_robot.parsers.base import BaseParser

logger = logging.getLogger("reactive_robot.connectors.base")


class Connector(ABC):
    """
    Base class for connectors.
    Connectors are for handling to event source configurations.
    """

    bindings: Iterable[BindingModel]
    variable_parser: BaseParser
    executor = ThreadPoolExecutor(max_workers=MAX_WORKER)

    def __init__(self):
        logger.info("Thread executor initialized with %s workers" % MAX_WORKER)

    @abstractmethod
    def bind(self, connection_url: ParseResult, bindings: Iterable[BindingModel]):
        raise NotImplemented



import logging
import subprocess
import tempfile
from typing import List
from urllib.parse import ParseResult

import paho.mqtt.client as mqtt

from reactive_robot.connectors.base import Connector
from reactive_robot.models import BindingModel
from reactive_robot.parsers.string import RawPayloadParser

logger = logging.getLogger("reactive_robot.connectors.mqtt")


def _run_job(variables: List[str], binding: BindingModel):
    variable_cli = ["-v " + "".join(var) for var in variables]
    with tempfile.TemporaryDirectory() as tmpdirname:
        cmd = " ".join(["robot",
                        "--outputdir", tmpdirname,
                        " ".join(variable_cli),
                        binding.robot.file])

        logger.info("Executing cmd, %s" % cmd)
        subprocess.run(cmd.split(" "))


def _find_binding_by_topic(topic_name: str, bindings: List[BindingModel]):
    for b in bindings:
        if b.topic == topic_name:
            return b


class MQTTConnector(mqtt.Client, Connector):

    def __init__(self, *args, **kwargs):
        super(MQTTConnector, self).__init__(*args, **kwargs)

    def on_connect(self, mqttc, obj, flags, rc):
        logger.info("rc: " + str(rc) + "rc")

    def on_connect_fail(self, mqttc, obj):
        logger.error("MQTT broker connection failed")

    def on_message(self, mqttc, obj, msg):
        logger.info(msg.topic + " " + str(msg.qos) + " " + str(msg.payload))
        binding = _find_binding_by_topic(msg.topic, self.bindings)
        parser = RawPayloadParser()
        self.executor.submit(_run_job, parser.get_variables(msg.payload), binding)

    def on_publish(self, mqttc, obj, mid):
        logger.debug("mid: " + str(mid))

    def on_subscribe(self, mqttc, obj, mid, granted_qos):
        logger.info("Subscribed: " + str(mid) + " " + str(granted_qos))

    def on_log(self, mqttc, obj, level, string):
        logger.debug(string)

    def bind(self, connection_url: ParseResult, bindings=None):
        if bindings is None:
            bindings = []

        self.bindings = bindings
        logger.info(connection_url.hostname)
        logger.info(connection_url.port)
        self.connect(connection_url.hostname, connection_url.port, 60)

        for b in bindings:
            logger.info("%s -> subscribing topic: %s" % (b.name, b.topic))
            self.subscribe(b.topic)

        rc = 0
        while rc == 0:
            rc = self.loop()
        return rc



from reactive_robot.parsers.base import BaseParser


class RawPayloadParser(BaseParser):

    def get_variables(self, message: bytes):
        if message:
            variables = message.decode("utf-8").split()
            return variables
        else:
            return []



import logging

from colorlog import ColoredFormatter


class Log(object):
    """
    Maintain logging level
    """
    format = '%(asctime)s - [%(levelname)s] - %(name)s::%(funcName)s::%(lineno)d - %(message)s'

    def __init__(self, log_name='reactive_robot', level=logging.INFO):
        self.logger = logging.getLogger(log_name)
        self.logger.setLevel(1)
        self.logger.propagate = False

        self.stream = logging.StreamHandler()
        self.stream.setLevel(level)
        self.stream.setFormatter(ColoredFormatter(self.format))

        self.stream.name = 'RxRobotStreamHandler'
        self.logger.addHandler(self.stream)



import logging
import os
import yaml
from contextlib import contextmanager

from marshmallow.exceptions import ValidationError

from reactive_robot.config.schema import RxRobotConfigSchema
from reactive_robot.exceptions import InvalidConfigurationFileException, ConfigurationFileNotExists

log = logging.getLogger('reactive_robot.config')


@contextmanager
def _open_config_file(config_file):
    """
    A context manager which yields an open file descriptor ready to be read.

    Accepts a filename as a string, an open or closed file descriptor, or None.
    When None, it defaults to `reactive-robot.yml` in the CWD. If a closed file descriptor
    is received, a new file descriptor is opened for the same file.

    The file descriptor is automatically closed when the context manager block is existed.
    """

    # Default to the standard config filename.
    if config_file is None:
        paths_to_try = ['reactive-robot.yml', 'reactive-robot.yaml']
    # If it is a string, we can assume it is a path and attempt to open it.
    elif isinstance(config_file, str):
        paths_to_try = [config_file]
    # If closed file descriptor, get file path to reopen later.
    elif getattr(config_file, 'closed', False):
        paths_to_try = [config_file.name]
    else:
        paths_to_try = None

    if paths_to_try:
        # config_file is not a file descriptor, so open it as a path.
        for path in paths_to_try:
            path = os.path.abspath(path)
            log.debug(f"Trying to load configuration file: {path}")
            try:
                config_file = open(path, 'rb')
                break
            except FileNotFoundError:
                log.info(f"Config file '{path}' does not exist.")
                continue
        else:
            log.error(f"Config file '{paths_to_try[0]}' does not exist.")
            raise ConfigurationFileNotExists(f"Config file '{paths_to_try[0]}' does not exist.")
    else:
        log.debug(f"Trying to load configuration file: {config_file}")
        config_file.seek(0)

    try:
        yield config_file
    finally:
        if hasattr(config_file, 'close'):
            config_file.close()


def load_config(config_file=None, **kwargs):
    """
    Load the configuration for a given file object or name

    The config_file can either be a file object, string or None. If it is None
    the default `reactive-robot.yml` filename will loaded.

    Extra kwargs are passed to the configuration to replace any default values
    unless they themselves are None.
    """
    options = kwargs.copy()

    # Filter None values from the options. This usually happens with optional
    # parameters from Click.
    for key, value in options.copy().items():
        if value is None:
            options.pop(key)

    config_schema = RxRobotConfigSchema()
    with _open_config_file(config_file) as fd:
        data = yaml.safe_load(fd)
        log.debug("Reading YAML file, %s " % data)
        try:
            dump = config_schema.load(data, unknown=True)
        except ValidationError as e:
            raise InvalidConfigurationFileException(e.messages)

    log.debug("Using schema %s" % dump)
    return dump



import asyncio
import logging
from importlib.metadata import import_module

from reactive_robot.exceptions import InjectionException
from reactive_robot.models import ReactiveRobotModel

logger = logging.getLogger("reactive_robot.serve")


def _check_robot_exists():
    try:
        robot = import_module("robot")
        logger.info("Using Robot Framework v%s" % robot.version.get_full_version())
    except ImportError:
        logger.critical("Robot Framework not installed in active Python interpreter")
        raise InjectionException("Robot Framework not installed")


def _get_connector(klass: str):
    try:
        module_name = klass.split(".")[:-1]
        cls_name = klass.split(".").pop()
        return getattr(import_module(".".join(module_name)), cls_name)  # get class from module
    except ImportError:
        err_msg = "Failed to inject %s class" % klass
        logger.critical(err_msg)
        raise InjectionException(err_msg)


def serve(config: ReactiveRobotModel):
    _check_robot_exists()

    loop = asyncio.get_event_loop()

    cls = _get_connector(config.connector.driver)
    connector = cls()
    coroutine = loop.run_in_executor(None, lambda: connector.bind(connection_url=config.connector.connection_url,
                                                                  bindings=config.bindings))

    logger.info(
        "%s Service Version %s has been started. Waiting for events." % (config.service_name, config.service_version))
    loop.run_until_complete(coroutine)

    loop.close()



import os
import sys

__author__ = "Yusuf Can Bayrak <yusufcanbayrak@gmail.com>"
__version__ = "0.1.0-alpha.1"

PYTHON_VERSION = f"{sys.version_info.major}.{sys.version_info.minor}"
PKG_DIR = os.path.dirname(os.path.abspath(__file__))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['reactive_robot',
 'reactive_robot.config',
 'reactive_robot.connectors',
 'reactive_robot.parsers']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'click>=8.0.3,<9.0.0',
 'colorlog>=6.6.0,<7.0.0',
 'importlib-metadata>=4.8.2,<5.0.0',
 'marshmallow>=3.14.1,<4.0.0',
 'minio>=7.1.1,<8.0.0',
 'paho-mqtt>=1.6.1,<2.0.0',
 'pika>=1.2.0,<2.0.0']

setup_kwargs = {
    'name': 'reactive-robot',
    'version': '0.1.0a1',
    'description': 'Fastest way to turn your robot workflows into event driven service.',
    'long_description': '# Reactive Robot (⚡ 🤖)\n\n<!-- ![pypi-badge](https://img.shields.io/pypi/v/reactive-robot) -->\n\n![stable](https://img.shields.io/static/v1?label=status&message=alpha-phase&color=yellow)\n\n## Mission\n\nThis project aims to turn Robot Framework projects into event-driven services using popular message brokers like RabbitMQ, Kafka or MQTT.\n\n## Usage\n\nCreate a definition file called `reactive-robot.yml` then paste following configuration;\n\n```yaml\n\nservice_name: Example Robot Service\nservice_version: 1.0.0\n\nconnector:\n  driver: reactive_robot.connectors.mqtt.MQTTConnector\n  connection_url: mqtt://localhost:1883\n\nbindings:\n  - name: Example Task\n    topic: robot-queue\n    robot:\n      file: your-robots/examples.robot\n      args: null\n```\n\nYou\'re all set!\nNow all you have to do is start the service;\n\n```\npython -m reactive_robot serve\n```\n\nYou should see the following output;\n\n```\n$ python -m reactive_robot serve\n2021-11-27 18:22:58,517 - [INFO] - reactive_robot.serve::serve::40 - Using Robot Framework v4.1.2 (Python 3.10.0 on darwin)\n2021-11-27 18:22:58,518 - [INFO] - reactive_robot.serve::serve::47 - Event loop started. Waiting for events.\n```\n\nFinally publish a message to see your robots are running.\n\n```\npython tests/mqtt/publish.py localhost 1883\n```\n## Examples\n\n### Robot Service with MQTT Broker\n\n```\ncd examples/mqtt\ndocker-compose up\n```\n\n## Recipes\n\nIn the [examples/](examples) directory you can find example projects which implements all recipes below;\n\n### Dockerize your service\n\nHere you can find an example Dockerfile to convert your Robot Framework projects into dockerized event-driven service \n\n```dockerfile\nFROM robotframework/rfdocker\n\nWORKDIR /opt/service\n\nCOPY . /opt/service\nRUN pip install -r requirements.txt  # Your project dependencies.\n\n# reactive-robot setup\nCOPY reactive-robot.yml .\nRUN pip install reactive-robot\n\nCMD ["python", "-m", "reactive-robot", "serve"]\n```\n\nThen, we can build the image with following;\n\n```\ndocker build -t robot-service:1.0.0 .\n```\n\nFinally, run your service;\n\n```\ndocker run -it robot-service:1.0.0\n```\n\n\n## License\n\nDistributed under the Apache License 2.\nSee `LICENSE` for more information.\n\n## Contributing\n\nContributions are what make the open source community such an amazing place to be learn, inspire, and create.\nAny contributions are **appreciated**.\n\n1. Fork the Project\n2. Create your Feature Branch (`git checkout -b feature/some-feature`)\n3. Commit your Changes (`git commit -m \'some feature added\'`)\n4. Push to the Branch (`git push origin feature/some-feature`)\n5. Open a Pull Request\n',
    'author': 'Yusuf Can Bayrak',
    'author_email': 'yusufcanbayrak@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
